from typing import Callable, Optional

try:
    from Crypto.Cipher import AES
    from Crypto.Util.Padding import pad, unpad
    from Crypto.Protocol.KDF import scrypt
except ImportError:
    raise ImportError("PyCryptodome is required to use this module -- pip install pycryptodome")

from .mock import MockTransformer


class AESTransformer(MockTransformer):
    def __init__(
            self,
            passphrase: str,
            static_names_mode: bool,
            salt: str = 'dictature',
            bytes_encoder: Optional[Callable[[bytes], str]] = None,
            bytes_decoder: Optional[Callable[[str], bytes]] = None
    ) -> None:
        """
        Create a new AES transformer
        :param passphrase: secret passphrase to encrypt/decrypt the data
        :param static_names_mode: if True, the transformer will use ECB mode instead of GCM (True decreases security, increases speed)
        :param salt: salt to use for the key derivation
        :param bytes_encoder: function to encode bytes to string, can be e.g. b64encode (default: hex)
        :param bytes_decoder: function to decode string to bytes, can be e.g. b64decode (default: hex)
        """
        self.__key = scrypt(passphrase, salt, 16, N=2 ** 14, r=8, p=1)
        self.__mode = AES.MODE_GCM if not static_names_mode else AES.MODE_ECB
        self.__static = static_names_mode
        self.__bytes_encoder = bytes_encoder if bytes_encoder else (lambda b: b.hex())
        self.__bytes_decoder = bytes_decoder if bytes_decoder else (lambda s: bytes.fromhex(s))

    def forward(self, text: str) -> str:
        cipher = self.__cipher()
        if self.__mode == AES.MODE_GCM:
            ciphertext, tag = cipher.encrypt_and_digest(pad(text.encode('utf8'), AES.block_size))
            data = (cipher.nonce + tag + ciphertext)
        else:
            data = cipher.encrypt(pad(text.encode('utf8'), AES.block_size))

        return self.__bytes_encoder(data)

    def backward(self, text: str) -> str:
        data = self.__bytes_decoder(text)
        if self.__mode == AES.MODE_GCM:
            nonce, tag, ciphertext = data[:16], data[16:32], data[32:]
            return unpad(self.__cipher(nonce=nonce).decrypt_and_verify(ciphertext, tag), AES.block_size).decode('utf8')
        else:
            return unpad(self.__cipher().decrypt(data), AES.block_size).decode('utf8')

    def __cipher(self, **kwargs) -> AES:
        # noinspection PyTypeChecker
        return AES.new(self.__key, self.__mode, **kwargs)

    @property
    def static(self) -> bool:
        return self.__static



import json
import pickle
from gzip import compress, decompress
from base64 import b64encode, b64decode
from random import choice
from typing import Optional, Dict, Any, Set, Iterator, Tuple

from .backend import DictatureBackendMock, ValueMode, Value
from .transformer import MockTransformer, PassthroughTransformer



class Dictature:
    def __init__(
            self,
            backend: DictatureBackendMock,
            name_transformer: MockTransformer = PassthroughTransformer(),
            value_transformer: MockTransformer = PassthroughTransformer(),
            table_name_transformer: Optional[MockTransformer] = None,
            allow_pickle: bool = False,
            allow_invalid_keys: bool = True,
    ) -> None:
        """
        Create a new Dictature object
        :param backend: backend to use
        :param name_transformer: transformer to use for table and key names
        :param value_transformer: transformer to use for values
        :param table_name_transformer: transformer to use for table names, if None, name_transformer is used
        :param allow_pickle: if True, pickle is allowed for values (warning: this may be a security risk if the data is not trusted)
        :param allow_invalid_keys: if True, keys that cannot be decoded are quietly skipped
        """
        self.__backend = backend
        self.__table_cache: Dict[str, "DictatureTable"] = {}
        self.__name_transformer = name_transformer
        self.__value_transformer = value_transformer
        self.__table_name_transformer = table_name_transformer or name_transformer
        self.__cache_size = 4096
        self.__allow_pickle = allow_pickle
        self.__allow_invalid_keys = allow_invalid_keys

    def keys(self) -> Set[str]:
        """
        Return all table names
        :return: all table names
        """
        result = set()
        for key in self.__backend.keys():
            try:
                decoded = self.__table_name_transformer.backward(key)
            except Exception:
                if self.__allow_invalid_keys:
                    continue
                raise
            result.add(decoded)
        return result

    def values(self) -> Iterator["DictatureTable"]:
        """
        Return all tables
        :return: all tables
        """
        return map(lambda x: x[1], self.items())

    def items(self) -> Iterator[Tuple[str, "DictatureTable"]]:
        """
        Return all tables with their instances
        :return: all tables with their instances
        """
        for k in self.keys():
            yield k, self[k]

    def to_dict(self) -> Dict[str, Any]:
        """
        Return all tables as a dictionary
        :return: all tables as a dictionary
        """
        return {k: v.to_dict() for k, v in self.items()}

    def __str__(self):
        """
        Return all tables as a string
        :return: all tables as a string
        """
        return str(self.to_dict())

    def __getitem__(self, item: str) -> "DictatureTable":
        """
        Get a table by name
        :param item: name of the table
        :return: table instance
        """
        if len(self.__table_cache) > self.__cache_size:
            del self.__table_cache[choice(list(self.__table_cache.keys()))]
        if item not in self.__table_cache:
            self.__table_cache[item] = DictatureTable(
                self.__backend,
                item,
                name_transformer=self.__name_transformer,
                value_transformer=self.__value_transformer,
                table_name_transformer=self.__table_name_transformer,
                allow_pickle=self.__allow_pickle,
                allow_invalid_keys=self.__allow_invalid_keys,
            )
        return self.__table_cache[item]

    def __delitem__(self, key: str) -> None:
        """
        Delete a table
        :param key: name of the table
        :return: None
        """
        self[key].drop()

    def __contains__(self, item: str) -> bool:
        """
        Check if a table exists
        :param item: name of the table
        :return: True if the table exists, False otherwise
        """
        return item in self.keys()

    def __bool__(self) -> bool:
        """
        Check if there are any tables
        :return: True if there are tables, False otherwise
        """
        return not not self.keys()


class DictatureTable:
    def __init__(
            self,
            backend: DictatureBackendMock,
            table_name: str,
            name_transformer: MockTransformer = PassthroughTransformer(),
            value_transformer: MockTransformer = PassthroughTransformer(),
            table_name_transformer: Optional[MockTransformer] = None,
            allow_pickle: bool = False,
            allow_invalid_keys: bool = True,
    ):
        """
        Create a new DictatureTable object
        :param backend: backend to use
        :param table_name: name of the table
        :param name_transformer:  transformer to use for key names
        :param value_transformer: transformer to use for values
        :param allow_pickle: if True, pickle is allowed for values (warning: this may be a security risk if the data is not trusted)
        :param allow_invalid_keys: if True, keys that cannot be decoded are quietly skipped
        """
        self.__backend = backend
        self.__name_transformer = name_transformer
        self.__value_transformer = value_transformer
        self.__table_name_transformer = table_name_transformer or name_transformer
        self.__table = self.__backend.table(self.__table_key(table_name))
        self.__table_created = False
        self.__allow_pickle = allow_pickle
        self.__allow_invalid_keys = allow_invalid_keys

    def get(self, item: str, default: Optional[Any] = None) -> Any:
        """
        Get a value from the table
        :param item: key to get
        :param default: default value to return if the key does not exist
        :return: value or default
        """
        try:
            return self[item]
        except KeyError:
            return default

    def key_exists(self, item: str) -> bool:
        """
        Check if a key exists
        :param item: key to check
        :return: True if the key exists, False otherwise
        """
        self.__create_table()
        return item in self.keys()

    def keys(self) -> Set[str]:
        """
        Return all keys in the table
        :return: all keys in the table
        """
        self.__create_table()
        result = set()
        for key in self.__table.keys():
            try:
                decoded = self.__name_transformer.backward(key)
            except Exception:
                if self.__allow_invalid_keys:
                    continue
                raise
            result.add(decoded)
        return result

    def values(self) -> Iterator[Any]:
        """
        Return all values in the table
        :return: all values in the table
        """
        return map(lambda x: x[1], self.items())

    def items(self) -> Iterator[Tuple[str, Any]]:
        """
        Return all items in the table
        :return: all items in the table
        """
        for k in self.keys():
            yield k, self[k]

    def drop(self) -> None:
        """
        Delete the table
        :return: None
        """
        self.__create_table()
        self.__table.drop()

    def to_dict(self) -> Dict[str, Any]:
        """
        Return all items as a dictionary
        :return: all items as a dictionary
        """
        return {k: v for k, v in self.items()}

    def __str__(self):
        """
        Return all items as a string
        :return: all items as a string
        """
        return str(self.to_dict())

    def __getitem__(self, item: str) -> Any:
        """
        Get a value from the table
        :param item: key to get
        :return: value
        """
        self.__create_table()
        saved_value = self.__table.get(self.__item_key(item))
        mode = ValueMode(saved_value.mode)
        value = self.__value_transformer.backward(saved_value.value)
        if mode == ValueMode.string:
            return value
        elif mode == ValueMode.json:
            return json.loads(value)
        elif mode == ValueMode.pickle:
            if not self.__allow_pickle:
                raise ValueError("Pickle is not allowed")
            return pickle.loads(decompress(b64decode(value.encode('ascii'))))
        raise ValueError(f"Unknown mode '{mode}'")

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set a value in the table
        :param key: key to set
        :param value: value to set
        :return: None
        """
        self.__create_table()
        value_mode: int = ValueMode.string.value

        if type(value) is not str:
            try:
                value = json.dumps(value)
                value_mode = ValueMode.json.value
            except TypeError:
                if not self.__allow_pickle:
                    raise ValueError("Pickle is not allowed")
                value = b64encode(compress(pickle.dumps(value))).decode('ascii')
                value_mode = ValueMode.pickle.value

        key = self.__item_key(key)
        value = self.__value_transformer.forward(value)
        self.__table.set(key, Value(value=value, mode=value_mode))

    def __delitem__(self, key: str) -> None:
        """
        Delete a key from the table
        :param key: key to delete
        :return: None
        """
        self.__table.delete(self.__item_key(key))

    def __contains__(self, item: str):
        """
        Check if a key exists
        :param item: key to check
        :return: True if the key exists, False otherwise
        """
        return item in self.keys()

    def __bool__(self) -> bool:
        """
        Check if there are any items in the table
        :return: True if there are items, False otherwise
        """
        return not not self.keys()

    def __create_table(self) -> None:
        """
        Create the table if it does not exist
        :return: None
        """
        if self.__table_created:
            return
        self.__table.create()
        self.__table_created = True

    def __item_key(self, item: str) -> str:
        """
        Transform the key for storage
        :param item: key to transform
        :return: transformed key
        """
        if not self.__name_transformer.static:
            for key in self.__table.keys():
                if self.__name_transformer.backward(key) == item:
                    return key
        return self.__name_transformer.forward(item)

    def __table_key(self, table_name: str) -> str:
        """
        Transform the table name for storage
        :param table_name: table name to transform
        :return: transformed table name
        """
        if not self.__table_name_transformer.static:
            for key in self.__backend.keys():
                if self.__table_name_transformer.backward(key) == table_name:
                    return key
        return self.__table_name_transformer.forward(table_name)



import gzip
import base64

from .mock import MockTransformer


class GzipTransformer(MockTransformer):
    """
    Compresses and decompresses text using Gzip.

    The compressed data is Base64 encoded to ensure it can be represented as a string.
    """
    def __init__(self) -> None:
        """
        Initializes the GzipTransformer. No parameters needed for basic Gzip.
        """
        # No specific state needed for standard gzip compression/decompression
        pass

    def forward(self, text: str) -> str:
        """
        Compresses the input text using Gzip and returns a Base64 encoded string.
        :param text: The original text string.
        :return: A Base64 encoded string representing the Gzipped data.
        """
        byte_data = text.encode('utf-8')
        compressed_bytes = gzip.compress(byte_data)
        base64_bytes = base64.b64encode(compressed_bytes)
        base64_string = base64_bytes.decode('ascii')
        return base64_string

    def backward(self, text: str) -> str:
        """
        Decompresses the Base64 encoded Gzip data back into the original text.
        :param text: A Base64 encoded string representing Gzipped data.
        :return: The original text string.
        :raises ValueError: If the input string is not valid Base64 or not valid Gzip data.
        """
        try:
            base64_bytes = text.encode('ascii')
            compressed_bytes = base64.b64decode(base64_bytes)
            original_bytes = gzip.decompress(compressed_bytes)
            original_text = original_bytes.decode('utf-8')
            return original_text
        except (gzip.BadGzipFile, UnicodeDecodeError) as e:
            # Catch errors related to Base64 decoding, Gzip decompression, or UTF-8 decoding
            raise ValueError(f"Invalid input data for Gzip decompression: {e}") from e

    @property
    def static(self) -> bool:
        return False





from .mock import MockTransformer


class PassthroughTransformer(MockTransformer):
    """
    Passthrough transformer, does not modify the text.
    """

    def forward(self, text: str) -> str:
        return text

    def backward(self, text: str) -> str:
        return text

    @property
    def static(self) -> bool:
        return True



from json import dumps, loads
from string import hexdigits, ascii_letters, digits, printable
from typing import Iterable, NamedTuple
from enum import Enum


class ValueMode(Enum):
    string = 0
    json = 1
    pickle = 2


class Value(NamedTuple):
    value: str
    mode: int


class ValueSerializerMode(Enum):
    any_string = 0
    ascii_only = 1
    filename_only = 2
    hex_only = 3


class ValueSerializer:
    prefix = 'a09e'

    def __init__(self, mode: ValueSerializerMode = ValueSerializerMode.any_string):
        self.__mode = mode

    def deserialize(self, serialized: str) -> Value:
        """
        Deserialize a string into a Value object
        :param serialized: serialized string
        :return: Value object
        """
        # Check if the string starts with the prefix (hex-encoded)
        if serialized.startswith(self.prefix):
            # Decode the hex part
            hex_data = serialized[len(self.prefix):]
            decoded = bytes.fromhex(hex_data).decode('ascii')
            # Recursively deserialize the decoded string
            return ValueSerializer(mode=ValueSerializerMode.ascii_only).deserialize(decoded)

        # Check if the string looks like JSON
        if serialized.startswith('{'):
            try:
                data = loads(serialized)
                return Value(value=data['value'], mode=data['mode'])
            except (ValueError, KeyError):
                pass  # Not valid JSON or missing required keys

        # Direct value (string mode)
        return Value(value=serialized, mode=ValueMode.string.value)

    def serialize(self, value: Value) -> str:
        """
        Serializes a `Value` object into a `str`, converting its data representation
        based on the serialization mode set in `ValueSerializerMode`. Depending on
        the mode provided, the object can be serialized directly if its string
        representation conforms to certain criteria, or it may be converted into
        a JSON format. Handles customization of allowed characters and uses prefix
        to encode conditions if the direct representation is not permitted.

        If the mode is incompatible or unsupported, raises a `NotImplementedError`.

        :param value: Instance of `Value` to be serialized.

        :return: Serialized representation of the `Value` object as a string.

        :raises NotImplementedError: If the mode provided in `ValueSerializerMode`
            is unsupported.
        """
        if self.__mode == ValueSerializerMode.hex_only:
            allowed_chars = hexdigits
        elif self.__mode == ValueSerializerMode.filename_only:
            allowed_chars = ascii_letters + digits + ' -_.'
        elif self.__mode in (ValueSerializerMode.any_string, ValueSerializerMode.ascii_only):
            allowed_chars = None
        else:
            raise NotImplementedError(self.__mode)

        if allowed_chars is not None:
            # Only a subset of characters is allowed if all match and do not start with the reserved prefix, encode directly
            if value.mode == ValueMode.string.value and all(map(lambda x: x in allowed_chars, value.value)) and not value.value.startswith(self.prefix):
                return value.value
            return self.prefix + ValueSerializer(mode=ValueSerializerMode.ascii_only).serialize(value).encode('ascii').hex()

        # Save as JSON if not string or value is starting with { (indicating JSON)
        save_as_json = value.mode != ValueMode.string.value or value.value.startswith('{') or value.value.startswith(self.prefix)
        # Save as JSON if only ASCII strings are allowed
        save_as_json = save_as_json or (self.__mode == ValueSerializerMode.ascii_only and any(filter(lambda x: x not in printable, value.value)))
        return dumps({'value': value.value, 'mode': value.mode}, indent=1) if save_as_json else value.value


class DictatureBackendMock:
    def keys(self) -> Iterable[str]:
        """
        Return all table names
        :return: all table names
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def table(self, name: str) -> 'DictatureTableMock':
        """
        Create a table object based on the name
        :param name: name of the table
        :return: table object
        """
        raise NotImplementedError("This method should be implemented by the subclass")


class DictatureTableMock:
    def keys(self) -> Iterable[str]:
        """
        Return all keys in the table
        :return: all keys in the table
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def drop(self) -> None:
        """
        Delete the table
        :return: None
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def create(self) -> None:
        """
        Create the table in the backend
        :return: None
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def set(self, item: str, value: Value) -> None:
        """
        Set a value in the table
        :param item: key to set
        :param value: value to set
        :return: None
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def get(self, item: str) -> Value:
        """
        Get a value from the table
        :param item: key to get
        :return: value
        :raises KeyError: if the key does not exist
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def delete(self, item: str) -> None:
        """
        Delete a value from the table
        :param item: key to delete
        :return: None
        """
        raise NotImplementedError("This method should be implemented by the subclass")


from typing import List

from .mock import MockTransformer


class PipelineTransformer(MockTransformer):
    def __init__(self, transformers: List[MockTransformer]) -> None:
        """
        Create a pipeline of transformers. The text is passed through each transformer in the order they are provided.
        :param transformers: list of transformers to use
        """
        self.__transformers = transformers

    def forward(self, text: str) -> str:
        for transformer in self.__transformers:
            text = transformer.forward(text)
        return text

    def backward(self, text: str) -> str:
        for transformer in reversed(self.__transformers):
            text = transformer.backward(text)
        return text

    @property
    def static(self) -> bool:
        return all(t.static for t in self.__transformers)



import io
import posixpath
from typing import Iterable

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    raise ImportError('Requires: pip install boto3')

from .mock import DictatureTableMock, DictatureBackendMock, Value, ValueMode, ValueSerializer, ValueSerializerMode


class DictatureBackendS3(DictatureBackendMock):
    def __init__(
            self,
            bucket_name: str,
            aws_access_key_id: str = None,
            aws_secret_access_key: str = None,
            region_name: str = None,
            endpoint_url: str = None,
            dir_prefix: str = 'db_',
            item_prefix: str = 'item_'
    ) -> None:
        """
        Initialize S3 backend

        :param bucket_name: S3 bucket name
        :param aws_access_key_id: AWS access key ID
        :param aws_secret_access_key: AWS secret access key
        :param region_name: AWS region name
        :param endpoint_url: Custom endpoint URL for alternative S3-compatible storage (e.g., MinIO)
        :param dir_prefix: Prefix for table directories
        :param item_prefix: Prefix for item files
        """
        self.__s3 = boto3.resource(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            endpoint_url=endpoint_url
        )
        self.__bucket = self.__s3.Bucket(bucket_name)
        self.__bucket_name = bucket_name
        self.__dir_prefix = dir_prefix
        self.__item_prefix = item_prefix

    def keys(self) -> Iterable[str]:
        """
        Get all table names in the S3 bucket
        """
        # Group objects by their "directory" prefix
        prefixes = set()
        for obj in self.__bucket.objects.filter(Prefix=self.__dir_prefix):
            key = obj.key
            parts = key.split('/')
            if len(parts) > 1 and parts[0].startswith(self.__dir_prefix):
                prefixes.add(parts[0])

        for prefix in prefixes:
            table_name_encoded = prefix[len(self.__dir_prefix):]
            # noinspection PyProtectedMember
            yield DictatureTableS3._filename_decode(table_name_encoded, suffix='')

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableS3(
            self.__s3,
            self.__bucket_name,
            name,
            self.__dir_prefix,
            self.__item_prefix
        )


class DictatureTableS3(DictatureTableMock):
    def __init__(
            self,
            s3_resource,
            bucket_name: str,
            name: str,
            db_prefix: str,
            prefix: str
    ) -> None:
        self.__s3 = s3_resource
        self.__bucket = self.__s3.Bucket(bucket_name)
        self.__bucket_name = bucket_name
        self.__encoded_name = self._filename_encode(name, suffix='')
        self.__path = f"{db_prefix}{self.__encoded_name}/"
        self.__prefix = prefix
        self.__name_serializer = ValueSerializer(mode=ValueSerializerMode.filename_only)
        self.__value_serializer = ValueSerializer(mode=ValueSerializerMode.any_string)

    def keys(self) -> Iterable[str]:
        try:
            for obj in self.__bucket.objects.filter(Prefix=self.__path):
                key = obj.key
                filename = posixpath.basename(key)
                if filename.startswith(self.__prefix):
                    item_name_encoded = filename[len(self.__prefix):]
                    yield self._filename_decode(item_name_encoded, suffix='.txt')
        except ClientError:
            pass  # Return empty generator if bucket doesn't exist or other error

    def drop(self) -> None:
        """
        Delete all objects in this table's directory
        """
        try:
            self.__bucket.objects.filter(Prefix=self.__path).delete()
        except ClientError:
            pass  # Ignore if objects don't exist or other error

    def create(self) -> None:
        """
        S3 doesn't need explicit directory creation, but we can create a marker file
        to indicate the table's existence if it doesn't already have items
        """
        try:
            # Check if the "directory" exists by looking for any objects with this prefix
            objects = list(self.__bucket.objects.filter(Prefix=self.__path).limit(1))
            if not objects:
                # Create an empty marker object to represent the directory
                self.__bucket.put_object(Key=f"{self.__path}.keep")
        except ClientError:
            # If bucket doesn't exist, this will fail, but that's expected
            pass

    def set(self, item: str, value: Value) -> None:
        item_path = self.__item_path(item)
        save_data: str = self.__value_serializer.serialize(value)
        data_bytes = save_data.encode('utf-8')

        with io.BytesIO(data_bytes) as buffer:
            self.__bucket.upload_fileobj(buffer, item_path)

    def get(self, item: str) -> Value:
        item_path = self.__item_path(item)
        buffer = io.BytesIO()
        try:
            self.__bucket.download_fileobj(item_path, buffer)
            buffer.seek(0)
            save_data = buffer.read().decode('utf-8')
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchKey':
                raise KeyError(item)
            raise  # Re-raise for other errors
        return self.__value_serializer.deserialize(save_data)

    def delete(self, item: str) -> None:
        """
        Delete the item from S3
        """
        item_path = self.__item_path(item)
        try:
            self.__s3.Object(self.__bucket_name, item_path).delete()
        except ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchKey':
                raise  # Only ignore "not found" errors

    def __item_path(self, item: str) -> str:
        encoded_item_name = self.__prefix + self._filename_encode(item, suffix='.txt')
        full_path = f"{self.__path}{encoded_item_name}"
        return full_path

    @staticmethod
    def _filename_encode(name: str, suffix: str = '.txt') -> str:
        return ValueSerializer(mode=ValueSerializerMode.filename_only).serialize(Value(
            value=name,
            mode=ValueMode.string.value
        )) + suffix

    @staticmethod
    def _filename_decode(name: str, suffix: str = '.txt') -> str:
        if suffix and name.endswith(suffix):
            name = name[:-len(suffix)]
        return ValueSerializer(mode=ValueSerializerMode.filename_only).deserialize(name).value



from .mock import DictatureBackendMock, ValueMode, Value
from .directory import DictatureBackendDirectory
from .sqlite import DictatureBackendSQLite
from .single_table import DictatureSingleTableBackend



import os
import unittest
from itertools import product
from typing import NamedTuple, Optional
from tempfile import mkdtemp, mktemp
from base64 import b64decode, b64encode

from parameterized import parameterized

from src.dictature import Dictature
from src.dictature.backend.mock import DictatureBackendMock
from src.dictature.backend.sqlite import DictatureBackendSQLite
from src.dictature.backend.directory import DictatureBackendDirectory
from src.dictature.backend.single_table import DictatureSingleTableBackend
from src.dictature.transformer import PassthroughTransformer, PipelineTransformer
from src.dictature.transformer.mock import MockTransformer
from src.dictature.transformer.aes import AESTransformer
from src.dictature.transformer.hmac import HmacTransformer
from src.dictature.transformer.gzip import GzipTransformer


def initialize_extra_backends() -> list:
    """Add optional backends that depend on external services."""
    backends = []

    # Add Baserow backend if credentials are available
    baserow_token = os.getenv('BASEROW_TOKEN')
    baserow_table_id = os.getenv('BASEROW_TABLE_ID')
    baserow_url = os.getenv('BASEROW_URL', 'https://api.baserow.io')
    if baserow_token and baserow_table_id:
        try:
            from src.dictature.backend.baserow import DictatureBackendBaserowSingleTable
            backends.append(DictatureBackendBaserowSingleTable(
                token=baserow_token,
                table_id=int(baserow_table_id),
                base_url=baserow_url
            ))
        except (ImportError, ValueError) as e:
            print(f"Warning: Could not load Baserow backend: {e}")
    else:
        print("Warning: BASEROW_TOKEN and BASEROW_TABLE_ID not set. Baserow backend will be skipped.")

    # Add MISP backend if credentials are available
    misp_url = os.getenv('MISP_URL')
    misp_key = os.getenv('MISP_KEY')
    if misp_url and misp_key:
        try:
            from pymisp import PyMISP
            from src.dictature.backend.misp import DictatureBackendMISP
            misp_instance = PyMISP(misp_url, misp_key, ssl=os.getenv('MISP_SSL', 'true').lower() == 'true')
            backends.append(DictatureBackendMISP(misp_instance))
        except ImportError as e:
            print(f"Warning: Could not load MISP backend: {e}")
    else:
        print("Warning: MISP_URL and MISP_KEY not set. MISP backend will be skipped.")

    # Add MySQL backend if credentials are available
    mysql_host = os.getenv('MYSQL_HOST')
    mysql_user = os.getenv('MYSQL_USER')
    mysql_password = os.getenv('MYSQL_PASSWORD')
    mysql_database = os.getenv('MYSQL_DATABASE')
    if mysql_host and mysql_user and mysql_password and mysql_database:
        try:
            from src.dictature.backend.mysql import DictatureBackendMySQL
            mysql_port = int(os.getenv('MYSQL_PORT', '3306'))
            backends.append(DictatureBackendMySQL(
                host=mysql_host,
                port=mysql_port,
                user=mysql_user,
                password=mysql_password,
                database=mysql_database
            ))
        except (ImportError, ValueError) as e:
            print(f"Warning: Could not load MySQL backend: {e}")
    else:
        print("Warning: MYSQL_HOST, MYSQL_USER, MYSQL_PASSWORD, and MYSQL_DATABASE not set. MySQL backend will be skipped.")

    # Add S3 backend if credentials are available
    s3_bucket = os.getenv('S3_BUCKET')
    if s3_bucket:
        try:
            from src.dictature.backend.s3 import DictatureBackendS3
            backends.append(DictatureBackendS3(
                bucket_name=s3_bucket,
                aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'],
                region_name=os.getenv('AWS_REGION', 'us-east-1'),
                endpoint_url=os.environ['S3_ENDPOINT_URL']
            ))
        except ImportError as e:
            print(f"Warning: Could not load S3 backend: {e}")
    else:
        print("Warning: S3_BUCKET not set. S3 backend will be skipped.")

    # Add WebDAV backend if credentials are available
    webdav_url = os.getenv('WEBDAV_URL')
    webdav_login = os.getenv('WEBDAV_LOGIN')
    webdav_password = os.getenv('WEBDAV_PASSWORD')
    if webdav_url and webdav_login and webdav_password:
        try:
            from webdav3.client import Client as WebdavClient
            from src.dictature.backend.webdav import DictatureBackendWebdav
            webdav_client = WebdavClient({
                'webdav_hostname': webdav_url,
                'webdav_login': webdav_login,
                'webdav_password': webdav_password,
                'disable_check': True,
            })
            backends.append(DictatureBackendWebdav(client=webdav_client))
        except ImportError as e:
            print(f"Warning: Could not load WebDAV backend: {e}")
    else:
        print("Warning: WEBDAV_URL, WEBDAV_LOGIN, and WEBDAV_PASSWORD not set. WebDAV backend will be skipped.")

    confluence_url = os.getenv('CONFLUENCE_URL')
    confluence_user = os.getenv('CONFLUENCE_USER')
    confluence_token = os.getenv('CONFLUENCE_TOKEN')
    confluence_root_page = os.getenv('CONFLUENCE_ROOT_PAGE_ID')
    confluence_sync = os.getenv('CONFLUENCE_SYNC_PAGE_CONTENT', 'false').lower() == 'true'
    if confluence_url and confluence_user and confluence_token and confluence_root_page:
        try:
            from src.dictature.backend.confluence import DictatureBackendConfluence
            backends.append(DictatureBackendConfluence(
                base_url=confluence_url,
                username=confluence_user,
                api_token=confluence_token,
                root_page_id=confluence_root_page,
                sync_page_content=confluence_sync
            ))
        except ImportError as e:
            print(f"Warning: Could not load Confluence backend: {e}")
    else:
        print("Warning: CONFLUENCE_URL, CONFLUENCE_USER, CONFLUENCE_TOKEN, and CONFLUENCE_ROOT_PAGE_ID not set. Confluence backend will be skipped.")

    return backends


DEFAULT_BACKENDS = [  # Default backends are tested with all transformer combinations
    DictatureBackendDirectory(mkdtemp(prefix='dictature')),
    DictatureBackendSQLite(mktemp(prefix='dictature', suffix='.sqlite3')),
    DictatureSingleTableBackend(Dictature(DictatureBackendSQLite(mktemp(prefix='dictature_single_table', suffix='.sqlite3')), allow_pickle=True)['table']),
]
EXTRA_BACKENDS = initialize_extra_backends()  # Extra backends tested only with passthrough transformers


TRANSFORMERS = [
    PassthroughTransformer(),
    AESTransformer('password', False),
    AESTransformer('password', True),
    AESTransformer('password', True, bytes_encoder=(lambda x: b64encode(x).decode('ascii')), bytes_decoder=(lambda x: b64decode(x.encode('ascii')))),
    HmacTransformer(),
    HmacTransformer('password'),
    GzipTransformer(),
    PipelineTransformer([HmacTransformer(), AESTransformer('password', False)]),
]


class Settings(NamedTuple):
    backend: DictatureBackendMock
    name_transformer: MockTransformer
    value_transformer: MockTransformer
    table_name_transformer: Optional[MockTransformer]


def get_transformer_name(transformer) -> str:
    """Get a readable name for a transformer."""
    if transformer is None:
        return "None"
    
    class_name = type(transformer).__name__
    
    # Handle specific transformer types with more details
    if class_name == "AESTransformer":
        # Check for base64 encoding by looking at attributes
        has_encoder = hasattr(transformer, 'bytes_encoder') and transformer.bytes_encoder is not None
        encoded = "_b64" if has_encoder else ""
        deterministic = "_det" if getattr(transformer, 'deterministic', False) else ""
        return f"AES{deterministic}{encoded}"
    elif class_name == "HmacTransformer":
        # Check if it has a password
        has_password = hasattr(transformer, '_key') and transformer._key != b''
        return "HMAC_pwd" if has_password else "HMAC"
    elif class_name == "PipelineTransformer":
        # Get names of transformers in the pipeline
        if hasattr(transformer, 'transformers'):
            pipeline_names = [get_transformer_name(t) for t in transformer.transformers]
            return f"Pipeline({'+'.join(pipeline_names)})"
        return "Pipeline"
    elif class_name == "PassthroughTransformer":
        return "Passthrough"
    elif class_name == "GzipTransformer":
        return "Gzip"
    
    return class_name


def get_backend_name(backend) -> str:
    """Get a readable name for a backend."""
    class_name = type(backend).__name__
    
    # Simplify common backend names
    if class_name.startswith("DictatureBackend"):
        return class_name.replace("DictatureBackend", "")
    
    return class_name


def get_name_func(testcase_func, param_num, param):
    """Generate readable test names from Settings."""
    settings = param.args[0]
    
    backend_name = get_backend_name(settings.backend)
    name_trans = get_transformer_name(settings.name_transformer)
    value_trans = get_transformer_name(settings.value_transformer)
    table_trans = get_transformer_name(settings.table_name_transformer)
    
    return f"{testcase_func.__name__}_{backend_name}__name_{name_trans}__value_{value_trans}__table_{table_trans}"


SETTINGS = [
    (Settings(backend, name_transformer, value_transformer, table_name_transformer),)
    for backend, name_transformer, value_transformer, table_name_transformer in product(DEFAULT_BACKENDS, TRANSFORMERS, TRANSFORMERS, [*TRANSFORMERS, None])
] + [
    (Settings(backend, PassthroughTransformer(), PassthroughTransformer(), PassthroughTransformer()),)
    for backend in EXTRA_BACKENDS
]


class TestOperations(unittest.TestCase):
    def setUp(self):
        self.backend = None

    def tearDown(self):
        if self.backend:
            for table in self.backend.keys():
                del self.backend[table]

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_simle_set_and_get(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer,
        )
        table = self.backend['table']
        table['key'] = 'value'
        table['key2'] = 'value2'
        table['key'] = 'value3'
        self.backend['table2']['key'] = 'value'
        self.assertEqual(table['key'], 'value3')
        self.assertEqual(table.keys(), {'key', 'key2'})
        self.assertEqual(self.backend.keys(), {'table', 'table2'})

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_traversal_set_and_get(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        table = self.backend[r'..\..\..']
        table[r'..\..\..'] = r'../..\../'
        table[r'../../../'] = r'../..\..\.'
        self.backend[r'../../../'][r'..\..\..'] = r'../../../'
        self.assertEqual(table[r'..\..\..'], r'../..\../')
        self.assertEqual(table.keys(), {r'..\..\..', r'../../../'})
        self.assertEqual(self.backend.keys(), {r'..\..\..', r'../../../'})

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_special_set_and_get(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        table = self.backend['✅']
        table['🔑'] = '⭐'
        table['🗝️'] = '✨'
        table['🔑'] = '🌟'
        self.backend['❌']['🔑'] = '🏚️'
        self.assertEqual(table['🔑'], '🌟')
        self.assertEqual(table.keys(), {'🔑', '🗝️'})
        self.assertEqual(self.backend.keys(), {'✅', '❌'})

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_saving_json_value(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        value = {'key': 'value'}
        self.backend['table']['key'] = value
        self.assertDictEqual(self.backend['table']['key'], value)
        self.backend['table']['key'] = 2
        self.assertEqual(self.backend['table']['key'], 2)

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_saving_pickle_value(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer,
            allow_pickle=True,
        )
        self.backend['table']['key'] = NamedTuple
        self.assertEqual(self.backend['table']['key'], NamedTuple)

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_pickle_disabled_by_default(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer,
        )
        with self.assertRaises(ValueError):
            self.backend['table']['key'] = NamedTuple

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_deletion_of_table_key(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        table = self.backend['table']
        table['key'] = 'value'
        table['key2'] = 'value2'
        del table['key']
        self.assertEqual({'key2'}, table.keys())

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_deletion_of_whole_table(self, settings: Settings):
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        self.backend['table2']['key'] = 'value'
        self.backend['table']['key'] = 'value'
        del self.backend['table']
        self.assertEqual(self.backend.keys(), {'table2'})

    @parameterized.expand(SETTINGS, name_func=get_name_func)
    def test_bulk_storage_and_retrieval(self, settings: Settings):
        """Test storing and retrieving 110 key-value pairs to verify pagination works correctly."""
        self.backend = Dictature(
            backend=settings.backend,
            name_transformer=settings.name_transformer,
            value_transformer=settings.value_transformer,
            table_name_transformer=settings.table_name_transformer
        )
        table = self.backend['bulk_test']
        
        # Store 110 key-value pairs
        num_items = 110
        expected_data = {f'key_{i}': f'value_{i}' for i in range(1, num_items + 1)}
        
        for key, value in expected_data.items():
            table[key] = value
        
        # Verify all keys are present
        stored_keys = table.keys()
        self.assertEqual(stored_keys, set(expected_data.keys()))
        
        # Verify all values are correct
        for key, expected_value in expected_data.items():
            self.assertEqual(table[key], expected_value)


if __name__ == '__main__':
    unittest.main()



from typing import Iterable, Optional, Dict, Any

try:
    import requests
except ImportError:
    raise ImportError('Requires: pip install requests') from None

from .mock import DictatureBackendMock, DictatureTableMock, Value


class DictatureBackendBaserowSingleTable(DictatureBackendMock):
    REQUIRED_COLUMNS = ('table', 'key', 'value', 'mode')

    def __init__(
            self,
            token: str,
            table_id: int,
            base_url: str = 'https://api.baserow.io'
    ) -> None:
        """
        Create a new Baserow backend using a single table
        :param token: Baserow API token
        :param table_id: Baserow table ID (the table must have 'table', 'key', 'value', 'mode' fields)
        :param base_url: Baserow API base URL
        """
        self.__base_url = base_url.rstrip('/')
        self.__table_id = table_id
        self.__headers = {
            'Authorization': f'Token {token}'
        }
        self.__columns_validated = False

    def _validate_columns(self) -> None:
        """Validate that all required columns exist in the Baserow table (lazy, runs once on first access)."""
        if self.__columns_validated:
            return
        
        response = requests.get(
            f"{self.__base_url}/api/database/fields/table/{self.__table_id}/",
            headers=self.__headers
        )
        if not response.ok:
            raise RuntimeError(f"Baserow API error {response.status_code}: {response.text}")
        
        fields = response.json()
        existing_columns = {field['name'] for field in fields}
        missing_columns = set(self.REQUIRED_COLUMNS) - existing_columns
        
        if missing_columns:
            raise ValueError(
                f"Baserow table {self.__table_id} is missing required columns: {', '.join(sorted(missing_columns))}. "
                f"Required columns are: {', '.join(self.REQUIRED_COLUMNS)}"
            )
        
        self.__columns_validated = True

    def keys(self) -> Iterable[str]:
        """Return all table names."""
        # Get distinct table names using pagination
        seen_tables = set()
        url = f"/api/database/rows/table/{self.__table_id}/"
        params = {'user_field_names': 'true', 'size': 200}
        
        while url:
            data = self._request('GET', url, params=params)
            for row in data.get('results', []):
                table = row.get('table')
                if table and table not in seen_tables:
                    seen_tables.add(table)
                    yield table
            url = data.get('next')
            params = None

    def table(self, name: str) -> 'DictatureTableMock':
        self._validate_columns()
        return DictatureTableBaserowSingleTable(self, name)

    def _request(self, method: str, url: str, params: Optional[Dict[str, Any]] = None,
                 json: Optional[Dict[str, Any]] = None) -> Any:
        # Handle both relative URLs and full URLs (from pagination 'next' field)
        if url.startswith('http://') or url.startswith('https://'):
            full_url = url
        else:
            full_url = f"{self.__base_url}{url}"
        response = requests.request(method, full_url, params=params, json=json, headers=self.__headers)
        if not response.ok:
            raise RuntimeError(f"Baserow API error {response.status_code}: {response.text}")
        if response.status_code == 204:
            return None
        return response.json()

    def _table_id(self) -> int:
        return self.__table_id


class DictatureTableBaserowSingleTable(DictatureTableMock):
    def __init__(self, parent: DictatureBackendBaserowSingleTable, table_name: str) -> None:
        self.__parent = parent
        self.__table_name = table_name

    def keys(self) -> Iterable[str]:
        """Return all keys in this table."""
        table_id = self.__parent._table_id()
        
        # Use filter to only get rows for this table
        url = f"/api/database/rows/table/{table_id}/"
        params = {
            'user_field_names': 'true',
            'filter__table__equal': self.__table_name,
            'size': 200
        }
        
        while url:
            data = self.__parent._request('GET', url, params=params)
            for row in data.get('results', []):
                key = row.get('key')
                if key:
                    yield str(key)
            url = data.get('next')
            params = None

    def drop(self) -> None:
        """Delete all rows for this table."""
        table_id = self.__parent._table_id()
        
        # Get all row IDs for this table
        url = f"/api/database/rows/table/{table_id}/"
        params = {
            'user_field_names': 'true',
            'filter__table__equal': self.__table_name,
            'size': 200
        }
        
        rows_to_delete = []
        while url:
            data = self.__parent._request('GET', url, params=params)
            for row in data.get('results', []):
                rows_to_delete.append(row['id'])
            url = data.get('next')
            params = None
        
        # Delete all rows
        for row_id in rows_to_delete:
            try:
                self.__parent._request('DELETE', f"/api/database/rows/table/{table_id}/{row_id}/")
            except RuntimeError:
                pass

    def create(self) -> None:
        """Ensure the backend table exists (already handled by parent)."""
        self.__parent._table_id()

    def set(self, item: str, value: Value) -> None:
        """Set a value for a key in this table."""
        table_id = self.__parent._table_id()
        
        # Try to find existing row using table and key filters
        row = self.__find_row(item)
        
        payload = {
            'table': self.__table_name,
            'key': item,
            'value': value.value,
            'mode': value.mode
        }
        
        if row is not None:
            # Update existing row
            self.__parent._request(
                'PATCH',
                f"/api/database/rows/table/{table_id}/{row['id']}/",
                params={'user_field_names': 'true'},
                json=payload
            )
        else:
            # Create new row
            self.__parent._request(
                'POST',
                f"/api/database/rows/table/{table_id}/",
                params={'user_field_names': 'true'},
                json=payload
            )

    def get(self, item: str) -> Value:
        """Get a value for a key in this table."""
        row = self.__find_row(item)
        if row is None:
            raise KeyError(item)
        
        value = row.get('value')
        mode = row.get('mode')
        return Value(value='' if value is None else str(value), mode=int(mode or 0))

    def delete(self, item: str) -> None:
        """Delete a key from this table."""
        table_id = self.__parent._table_id()
        
        row = self.__find_row(item)
        if row is None:
            return
        
        self.__parent._request('DELETE', f"/api/database/rows/table/{table_id}/{row['id']}/")

    def __find_row(self, item: str) -> Optional[Dict[str, Any]]:
        """Find a row by table and key using backend filters."""
        table_id = self.__parent._table_id()
        
        data = self.__parent._request(
            'GET',
            f"/api/database/rows/table/{table_id}/",
            params={
                'user_field_names': 'true',
                'filter__table__equal': self.__table_name,
                'filter__key__equal': item,
                'size': 1
            }
        )
        results = data.get('results', [])
        return results[0] if results else None



from typing import Iterable, Optional

from .mock import DictatureTableMock, DictatureBackendMock, Value, ValueMode, ValueSerializer, ValueSerializerMode

try:
    from pymisp import PyMISP, MISPEvent, MISPAttribute, MISPTag
except ImportError as e:
    raise ImportError("Please install the 'pymisp' package to use the 'DictatureBackendMISP' backend.") from e


class DictatureBackendMISP(DictatureBackendMock):
    def __init__(self, misp: PyMISP, tag_name: str = 'storage:dictature', prefix: str = 'Dictature storage: ') -> None:
        """
        Create a new MISP backend
        :param misp: PyMISP instance
        :param tag_name: tag name to use for the tables
        :param prefix: prefix for the event names
        """
        self.__misp = misp
        self.__tag_name = tag_name
        self.__prefix = prefix

    def keys(self) -> Iterable[str]:
        for event in self.__misp.search(tags=[self.__tag_name], pythonify=True):
            name = event.info
            if not name.startswith(self.__prefix):
                continue
            yield name[len(self.__prefix):]

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableMISP(self.__misp, self.__prefix + name, self.__tag_name)


class DictatureTableMISP(DictatureTableMock):
    def __init__(self, misp: PyMISP, event_description: str, tag: str) -> None:
        self.__misp = misp
        self.__event_description = event_description
        self.__tag = tag
        self.__event: Optional[MISPEvent] = None
        self.__serializer = ValueSerializer(mode=ValueSerializerMode.ascii_only)

    def keys(self) -> Iterable[str]:
        for attribute in self.__event_attributes():
            # Deserialize the key to get the original value
            deserialized_key = self.__serializer.deserialize(attribute.value)
            yield deserialized_key.value

    def drop(self) -> None:
        self.__misp.delete_event(self.__get_event())

    def create(self) -> None:
        self.__get_event()

    def set(self, item: str, value: Value) -> None:
        item_name = self.__serializer.serialize(Value(value=item, mode=ValueMode.string.value))
        save_data = self.__serializer.serialize(value)

        for attribute in self.__event_attributes():
            if attribute.value == item_name:
                attribute.value = item_name
                attribute.comment = save_data
                self.__misp.update_attribute(attribute)
                break
        else:
            attribute = MISPAttribute()
            attribute.value = item_name
            attribute.comment = save_data
            attribute.type = 'comment'
            attribute.to_ids = False
            attribute.disable_correlation = True
            self.__misp.add_attribute(self.__get_event(), attribute)
            self.__get_event().attributes.append(attribute)

    def get(self, item: str) -> Value:
        item_name = self.__serializer.serialize(Value(value=item, mode=ValueMode.string.value))
        for attribute in self.__event_attributes():
            if attribute.value == item_name:
                return self.__serializer.deserialize(attribute.comment)
        raise KeyError(item)

    def delete(self, item: str) -> None:
        item_name = self.__serializer.serialize(Value(value=item, mode=ValueMode.string.value))
        for attribute in self.__event_attributes():
            if attribute.value == item_name:
                # First update the attribute as deletion is not recognized immediately
                attribute.type = 'other'
                self.__misp.update_attribute(attribute)
                self.__misp.delete_attribute(attribute)
                break

    def __get_event(self) -> MISPEvent:
        if self.__event is None:
            for event in self.__misp.search(tags=[self.__tag], eventinfo=self.__event_description, pythonify=True):
                if event.info == self.__event_description:
                    self.__event = event
                    break
            else:
                # Check if tag exists, create if not
                existing_tags = self.__misp.search_tags(self.__tag, strict_tagname=True, pythonify=True)
                if not type(existing_tags) is list or not existing_tags:
                    tag = MISPTag()
                    tag.name = self.__tag
                    tag.colour = '#000000'
                    tag.exportable = True
                    self.__misp.add_tag(tag)

                event = MISPEvent()
                event.info = self.__event_description
                event.distribution = 0
                event.threat_level_id = 4
                event.analysis = 0
                event.add_tag(self.__tag)
                self.__misp.add_event(event)
                self.__event = event
        return self.__event

    def __event_attributes(self) -> Iterable[MISPAttribute]:
        for attribute in self.__get_event().attributes:
            if attribute.type != 'comment' or (hasattr(attribute, 'deleted') and attribute.deleted):
                continue
            yield attribute



import hmac
from hashlib import sha256
from .mock import MockTransformer


class HmacTransformer(MockTransformer):
    def __init__(self, secret: str = 'dictature') -> None:
        """
        Perform HMAC on the text.
        :param secret: secret key to use for HMAC, if not provided works as a simple hash function
        """
        self.__secret = secret

    def forward(self, text: str) -> str:
        return f"{self.__hmac(text)}-{text}"

    def backward(self, text: str) -> str:
        mac, text = text.split('-', 1)
        if mac != self.__hmac(text):
            raise ValueError('Invalid HMAC')
        return text

    def __hmac(self, text: str) -> str:
        return hmac.new(self.__secret.encode('utf8'), text.encode('utf8'), sha256).hexdigest()

    @property
    def static(self) -> bool:
        return True




class MockTransformer:
    def forward(self, text: str) -> str:
        """
        Transform the text in some way to the data format in data storage
        :param text: text to transform
        :return: transformed text
        """
        raise NotImplementedError("This method should be implemented by the child class")

    def backward(self, text: str) -> str:
        """
        Transform the data format in data storage to the text
        :param text: text to transform
        :return: original text
        """
        raise NotImplementedError("This method should be implemented by the child class")

    @property
    def static(self) -> bool:
        """
        Returns True only if when the forward transformation is applied to the same text, the result is always the same
        :return: True if the transformation is static
        """
        raise NotImplementedError("This method should be implemented by the child class")



import sqlite3
from pathlib import Path
from typing import Iterable, Union, Optional

from .mock import DictatureBackendMock, DictatureTableMock, Value, ValueMode


class DictatureBackendSQLite(DictatureBackendMock):
    def __init__(self, file: Union[str, Path], prefix: str = 'tb_') -> None:
        """
        Create a new SQLite backend
        :param file: file to store the database
        :param prefix: prefix for the tables (default: 'tb_')
        """
        if isinstance(file, str):
            file = Path(file)
        self.__file = file
        self.__connection = sqlite3.connect(
            f"{self.__file.absolute()}",
            check_same_thread=False if sqlite3.threadsafety >= 3 else True
        )
        self.__cursor = self.__connection.cursor()
        self.__prefix = prefix.replace("'", "").replace('`', '')

    def keys(self) -> Iterable[str]:
        tables = self._execute(f"SELECT tbl_name FROM sqlite_master WHERE type='table' AND tbl_name LIKE '{self.__prefix}%'")
        return {table[0][len(self.__prefix):] for table in tables}

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableSQLite(self, name, self.__prefix)

    def _execute(self, query: str, data: tuple = ()) -> list:
        return list(self.__cursor.execute(query, data))

    def _commit(self) -> None:
        self.__connection.commit()

    def __del__(self):
        self.__connection.close()


class DictatureTableSQLite(DictatureTableMock):
    def __init__(self, parent: "DictatureBackendSQLite", name: str, prefix: str) -> None:
        self.__parent = parent
        self.__supports_jsonization: Optional[bool] = None
        self.__table = "`%s`" % (prefix + name).replace('`', '``')

    def keys(self) -> Iterable[str]:
        # noinspection PyProtectedMember
        return set(map(lambda x: x[0], self.__parent._execute(f"SELECT key FROM {self.__table}")))

    def drop(self) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(f"DROP TABLE {self.__table}")

    def key_exists(self, item: str) -> bool:
        # noinspection PyProtectedMember
        return not not self.__parent._execute(f"SELECT value FROM {self.__table} WHERE key=?", (item,))

    def create(self) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(f"""
        CREATE TABLE IF NOT EXISTS {self.__table} (
        `key`	TEXT NOT NULL UNIQUE,
        `value`	TEXT
        );
        """)

    def set(self, item: str, value: Value) -> None:
        if value.mode != ValueMode.string:
            self.__support_jsonization()

        if self.key_exists(item):
            if self.__test_supports_jsonization():
                # noinspection PyProtectedMember
                self.__parent._execute(f"""
                UPDATE {self.__table} SET value=?, jsonized=? WHERE key=?
                """, (value.value, value.mode, item))
            else:
                # noinspection PyProtectedMember
                self.__parent._execute(f"UPDATE {self.__table} SET value=? WHERE key=?", (value.value, item))
        else:
            if self.__test_supports_jsonization():
                # noinspection PyProtectedMember
                self.__parent._execute(
                    f"INSERT INTO {self.__table} (`key`,`value`,`jsonized`) VALUES (?,?,?);",
                    (item, value.value, value.mode)
                )
            else:
                # noinspection PyProtectedMember
                self.__parent._execute(
                    f"INSERT INTO {self.__table} (`key`,`value`) VALUES (?,?);", (item, value.value)
                )
        # noinspection PyProtectedMember
        self.__parent._commit()

    def get(self, item: str) -> Value:
        if self.__test_supports_jsonization():
            cmd = f"SELECT value, jsonized FROM {self.__table} WHERE key=?"
        else:
            cmd = f"SELECT value FROM {self.__table} WHERE key=?"

        # noinspection PyProtectedMember
        r = self.__parent._execute(cmd, (item,))
        if r:
            value: str = r[0][0]

            if self.__supports_jsonization:
                jsonized_state: int = r[0][1]
            else:
                jsonized_state = ValueMode.string.value

            return Value(value=value, mode=jsonized_state)
        raise KeyError(item)

    def delete(self, item: str) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(
            f"DELETE FROM {self.__table} WHERE `key`=?;", (item,)
        )
        # noinspection PyProtectedMember
        self.__parent._commit()

    def __test_supports_jsonization(self) -> bool:
        if self.__supports_jsonization is not None:
            return self.__supports_jsonization
        # noinspection PyProtectedMember
        for _, column_name, _, _, _, _ in self.__parent._execute(f"PRAGMA table_info({self.__table})"):
            if column_name == 'jsonized':
                self.__supports_jsonization = True
                return True
        self.__supports_jsonization = False
        return False

    def __support_jsonization(self) -> None:
        if self.__supports_jsonization or self.__test_supports_jsonization():
            return
        # noinspection PyProtectedMember
        self.__parent._execute(f"""
        ALTER TABLE {self.__table} ADD COLUMN jsonized INTEGER NOT NULL DEFAULT 0
        """)
        self.__supports_jsonization = True


import tempfile
import os
from pathlib import Path
from typing import Iterable, Union
from shutil import rmtree

from .mock import DictatureTableMock, DictatureBackendMock, Value, ValueMode, ValueSerializer, ValueSerializerMode


class DictatureBackendDirectory(DictatureBackendMock):
    def __init__(self, directory: Union[Path, str], dir_prefix: str = 'db_', item_prefix: str = 'item_') -> None:
        """
        Create a new directory backend
        :param directory: directory to store the data
        :param dir_prefix: prefix for the directories of the tables
        """
        if isinstance(directory, str):
            directory = Path(directory)
        self.__directory = directory
        self.__dir_prefix = dir_prefix
        self.__item_prefix = item_prefix

    def keys(self) -> Iterable[str]:
        for child in self.__directory.iterdir():
            if child.is_dir() and child.name.startswith(self.__dir_prefix):
                # noinspection PyProtectedMember
                yield DictatureTableDirectory._filename_decode(child.name[len(self.__dir_prefix):], suffix='')

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableDirectory(self.__directory, name, self.__dir_prefix, self.__item_prefix)


class DictatureTableDirectory(DictatureTableMock):
    def __init__(self, path_root: Path, name: str, db_prefix: str, prefix: str) -> None:
        self.__path = path_root / (db_prefix + self._filename_encode(name, suffix=''))
        self.__prefix = prefix
        self.__name_serializer = ValueSerializer(mode=ValueSerializerMode.filename_only)
        self.__value_serializer = ValueSerializer(mode=ValueSerializerMode.any_string)

    def keys(self) -> Iterable[str]:
        for child in self.__path.iterdir():
            if child.is_file() and child.name.startswith(self.__prefix) and not child.name.endswith('.tmp'):
                yield self._filename_decode(child.name[len(self.__prefix):])

    def drop(self) -> None:
        rmtree(self.__path)

    def create(self) -> None:
        self.__path.mkdir(parents=True, exist_ok=True)

    def set(self, item: str, value: Value) -> None:
        file_target = self.__item_path(item)

        # Create temporary file for atomic write
        handle, file_target_tmp_path = tempfile.mkstemp(prefix=file_target.name, suffix='.tmp', dir=file_target.parent)
        os.close(handle)
        file_target_tmp = Path(file_target_tmp_path)

        save_data = self.__value_serializer.serialize(value)

        file_target_tmp.write_text(save_data)
        file_target_tmp.rename(file_target)

    def get(self, item: str) -> Value:
        try:
            save_data = self.__item_path(item).read_text()
            return self.__value_serializer.deserialize(save_data)
        except FileNotFoundError:
            raise KeyError(item)

    def delete(self, item: str) -> None:
        if self.__item_path(item).exists():
            self.__item_path(item).unlink()

    def __item_path(self, item: str) -> Path:
        return self.__path / (self.__prefix + self._filename_encode(item))

    @staticmethod
    def _filename_encode(name: str, suffix: str = '.txt') -> str:
        return ValueSerializer(mode=ValueSerializerMode.filename_only).serialize(Value(
            value=name,
            mode=ValueMode.string.value
        )) + suffix

    @staticmethod
    def _filename_decode(name: str, suffix: str = '.txt') -> str:
        if suffix:
            name = name[:-len(suffix)]
        return ValueSerializer(mode=ValueSerializerMode.filename_only).deserialize(name).value



try:
    import mysql.connector
except ImportError:
    raise ImportError('Requires: pip install mysql-connector-python') from None
from typing import Iterable

from .mock import DictatureBackendMock, DictatureTableMock, Value


class DictatureBackendMySQL(DictatureBackendMock):
    def __init__(self, host: str, port: int = 3306, user: str = None, password: str = None,
                 database: str = None, prefix: str = 'tb_', **kwargs) -> None:
        """
        Create a new MySQL backend
        :param host: MySQL server host
        :param port: MySQL server port (default: 3306)
        :param user: MySQL username
        :param password: MySQL password
        :param database: MySQL database name
        :param kwargs: Additional connection parameters
        """
        self.__connection_params = {
            'host': host,
            'port': port,
            'user': user,
            'password': password,
            'database': database,
            **kwargs
        }

        # Remove None values from connection params
        self.__connection_params = {k: v for k, v in self.__connection_params.items() if v is not None}

        self._connect()
        self.__prefix = prefix.replace('`', '').replace("'", '')

    def _connect(self):
        self.__connection = mysql.connector.connect(**self.__connection_params)
        self.__cursor = self.__connection.cursor()

    def keys(self) -> Iterable[str]:
        # noinspection SqlResolve
        tables = self._execute(f"SELECT table_name FROM information_schema.tables WHERE table_schema = %s AND table_name LIKE '{self.__prefix}%'", (self.__connection_params['database'],))
        return {table[0][len(self.__prefix):] for table in tables}

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableMySQL(self, name, self.__prefix)

    def _execute(self, query: str, data: tuple = ()) -> list:
        try:
            self.__cursor.execute(query, data)
            return self.__cursor.fetchall()
        except mysql.connector.errors.Error as e:
            # Check for lost connection errors:
            # 2013: Lost connection to MySQL server during query
            # 2006: MySQL server has gone away
            # 4031: The client was disconnected by the server because of inactivity
            # 2055: Lost connection to MySQL server at '...'
            if e.errno in (2013, 2006, 2055, 4031):
                self._connect()
                self.__cursor.execute(query, data)
                return self.__cursor.fetchall()
            raise

    def _commit(self) -> None:
        self.__connection.commit()

    def __del__(self):
        if hasattr(self, '_DictatureBackendMySQL__connection'):
            try:
                self.__connection.close()
            except Exception:
                pass


class DictatureTableMySQL(DictatureTableMock):
    def __init__(self, parent: "DictatureBackendMySQL", name: str, prefix: str) -> None:
        self.__parent = parent
        # MySQL table names don't need backticks for simple names, but we'll use them for consistency
        self.__table = f"`{prefix}{name.replace('`', '')}`"

    def keys(self) -> Iterable[str]:
        # noinspection PyProtectedMember
        result = self.__parent._execute(f"SELECT `key` FROM {self.__table}")
        return set(map(lambda x: x[0], result))

    def drop(self) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(f"DROP TABLE {self.__table}")

    def key_exists(self, item: str) -> bool:
        # noinspection PyProtectedMember
        result = self.__parent._execute(f"SELECT `value` FROM {self.__table} WHERE `key`=%s", (item,))
        return len(result) > 0

    def create(self) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(f"""
        CREATE TABLE IF NOT EXISTS {self.__table} (
        `key` VARCHAR(700) NOT NULL UNIQUE,
        `value` TEXT,
        `type` INT NOT NULL DEFAULT 0,
        PRIMARY KEY (`key`)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_bin
        """)

    def set(self, item: str, value: Value) -> None:
        if self.key_exists(item):
            # noinspection PyProtectedMember
            self.__parent._execute(f"""
            UPDATE {self.__table} SET `value`=%s, `type`=%s WHERE `key`=%s
            """, (value.value, value.mode, item))
        else:
            # noinspection PyProtectedMember
            self.__parent._execute(
                f"INSERT INTO {self.__table} (`key`, `value`, `type`) VALUES (%s, %s, %s)",
                (item, value.value, value.mode)
            )
        # noinspection PyProtectedMember
        self.__parent._commit()

    def get(self, item: str) -> Value:
        # noinspection PyProtectedMember
        r = self.__parent._execute(f"SELECT `value`, `type` FROM {self.__table} WHERE `key`=%s", (item,))
        if r:
            value: str = r[0][0]
            type_value: int = r[0][1]
            return Value(value=value, mode=type_value)
        raise KeyError(item)

    def delete(self, item: str) -> None:
        # noinspection PyProtectedMember
        self.__parent._execute(
            f"DELETE FROM {self.__table} WHERE `key`=%s", (item,)
        )
        # noinspection PyProtectedMember
        self.__parent._commit()



"""Confluence backend using Content Properties for atomic KV storage."""
import re
import html
import time
from typing import Iterable, Optional, Dict, Any

try:
    import requests
except ImportError:
    raise ImportError('Requires: pip install requests') from None

from .mock import DictatureBackendMock, DictatureTableMock, Value, ValueSerializer, ValueSerializerMode


class DictatureBackendConfluence(DictatureBackendMock):
    """
    Backend that stores data in Confluence using Content Properties for atomicity.
    
    Each table is a child page under the root page. Key-value pairs are stored
    as Content Properties on the page resource.
    
    This allows atomic updates of individual keys.
    """

    def __init__(
            self,
            base_url: str,
            username: str,
            api_token: str,
            root_page_id: str,
            sync_page_content: bool = False,
    ) -> None:
        """
        Create a new Confluence backend.
        
        :param base_url: Confluence base URL (e.g., 'https://yoursite.atlassian.net/wiki')
        :param username: Atlassian account email/username
        :param api_token: Atlassian API token
        :param root_page_id: ID of the root page under which table pages will be created
        :param sync_page_content: If True, also update the page body with a table of all properties (slower)
        """
        self.__base_url = base_url.rstrip('/')
        self.__root_page_id = root_page_id
        self.__sync_page_content = sync_page_content
        self.__session = requests.Session()
        self.__session.auth = (username, api_token)
        self.__session.headers.update({
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        })
        # Cache the space key from root page
        self.__space_key: Optional[str] = None

    def _get_space_key(self) -> str:
        """Get the space key from the root page (cached)."""
        if self.__space_key is None:
            page = self._request('GET', f'/rest/api/content/{self.__root_page_id}', params={'expand': 'space'})
            self.__space_key = page['space']['key']
        return self.__space_key

    def keys(self) -> Iterable[str]:
        """Return all table names (child page titles under root page)."""
        url = f'/rest/api/content/{self.__root_page_id}/child/page'
        params = {'limit': 100}
        
        while url:
            data = self._request('GET', url, params=params)
            for page in data.get('results', []):
                yield page['title']
            
            # Handle pagination
            next_link = data.get('_links', {}).get('next')
            if next_link:
                url = next_link
                params = None
            else:
                url = None

    def table(self, name: str) -> 'DictatureTableConfluence':
        return DictatureTableConfluence(self, name, self.__sync_page_content)

    def _request(self, method: str, url: str, params: Optional[Dict[str, Any]] = None,
                 json: Optional[Dict[str, Any]] = None) -> Any:
        """Make an API request to Confluence."""
        if url.startswith('http://') or url.startswith('https://'):
            full_url = url
        else:
            full_url = f"{self.__base_url}{url}"
        
        response = self.__session.request(method, full_url, params=params, json=json)
        
        # Handle 404 (Not Found) gracefully if needed by caller, otherwise raise
        if response.status_code == 404:
            return None
            
        if not response.ok:
            # Pass through 409 Conflict for retry logic handling
            if response.status_code == 409:
                response.raise_for_status() 
            raise RuntimeError(f"Confluence API error {response.status_code}: {response.text}")
            
        if response.status_code == 204:
            return None
        
        try:
            return response.json()
        except ValueError:
            return None

    def _get_root_page_id(self) -> str:
        return self.__root_page_id


class DictatureTableConfluence(DictatureTableMock):
    """A table stored as a Confluence page using Content Properties."""

    MAX_RETRIES = 5

    def __init__(self, parent: DictatureBackendConfluence, table_name: str, sync_page_content: bool) -> None:
        self.__parent = parent
        self.__table_name = table_name
        self.__sync_page_content = sync_page_content
        self.__serializer = ValueSerializer(mode=ValueSerializerMode.any_string)

    def _find_page_id(self) -> Optional[str]:
        """Find the page ID for this table by title."""
        root_id = self.__parent._get_root_page_id()
        url = f'/rest/api/content/{root_id}/child/page'
        params = {'limit': 100}
        
        while url:
            data = self.__parent._request('GET', url, params=params)
            if not data:
                return None
                
            for page in data.get('results', []):
                if page['title'] == self.__table_name:
                    return page['id']
            
            next_link = data.get('_links', {}).get('next')
            if next_link:
                url = next_link
                params = None
            else:
                url = None
        
        return None

    def _encode_key(self, key: str) -> str:
        """Hex encode the key to meet content property key constraints."""
        return key.encode('utf-8').hex()

    def _decode_key(self, hex_key: str) -> str:
        """Decode hex key back to string."""
        return bytes.fromhex(hex_key).decode('utf-8')

    def create(self) -> str:
        """Create the table page if it doesn't exist, return page ID."""
        page_id = self._find_page_id()
        if page_id is not None:
            return page_id
        
        space_key = self.__parent._get_space_key()
        root_id = self.__parent._get_root_page_id()
        
        payload = {
            'type': 'page',
            'title': self.__table_name,
            'space': {'key': space_key},
            'ancestors': [{'id': root_id}],
            'body': {
                'storage': {
                    'value': '<p>This page acts as a key-value store database table.</p>',
                    'representation': 'storage',
                }
            }
        }
        
        try:
            data = self.__parent._request('POST', '/rest/api/content', json=payload)
            return data['id']
        except RuntimeError as e:
            # Handle race condition where page might be created by another process
            if "already exists" in str(e) or "400" in str(e): # Confluence often returns 400 for duplicate title
                 page_id = self._find_page_id()
                 if page_id:
                     return page_id
            raise e

    def keys(self) -> Iterable[str]:
        """Return all keys in this table (from content properties)."""
        page_id = self._find_page_id()
        if page_id is None:
            return
            
        url = f'/rest/api/content/{page_id}/property'
        params = {'limit': 100}
        
        while url:
            data = self.__parent._request('GET', url, params=params)
            if not data:
                break
                
            for prop in data.get('results', []):
                hex_key = prop['key']
                # Check if it's our property (we assume all properties on this page are ours)
                # But we should be careful about other plugins. 
                # Our keys are hex strings, so we can try to decode.
                try:
                    yield self._decode_key(hex_key)
                except ValueError:
                    pass # Ignore properties that aren't hex encoded (not ours)
            
            next_link = data.get('_links', {}).get('next')
            if next_link:
                url = next_link
                params = None
            else:
                url = None

    def set(self, item: str, value: Value) -> None:
        """Set a value for a key as a content property."""
        page_id = self.create()
        hex_key = self._encode_key(item)
        serialized = self.__serializer.serialize(value)
        
        self.__set_property(page_id, hex_key, serialized, value.mode, item)
        
        if self.__sync_page_content:
            self.__sync_content(page_id)

    def __set_property(self, page_id: str, hex_key: str, serialized_value: str, mode: int, original_key: str) -> None:
        """Set content property with optimistic locking for the property resource."""
        property_data = {
            "key": original_key,
            "value": serialized_value,
            "mode": mode
        }
        
        # Retry loop for property update
        for attempt in range(self.MAX_RETRIES):
            # 1. Try to get existing property to know version
            prop = self.__parent._request('GET', f'/rest/api/content/{page_id}/property/{hex_key}')
            
            if prop:
                # Update existing
                version = prop['version']['number'] + 1
                prop_id = prop['id']
                url = f'/rest/api/content/{page_id}/property/{hex_key}'
                payload = {
                    'version': {'number': version},
                    'value': property_data
                }
                method = 'PUT'
            else:
                # Create new
                url = f'/rest/api/content/{page_id}/property'
                payload = {
                    'key': hex_key,
                    'value': property_data
                }
                method = 'POST'
            
            try:
                self.__parent._request(method, url, json=payload)
                return # Success
            except requests.exceptions.HTTPError as e: # requests raises HTTPError for 409 if we use raise_for_status
                if e.response.status_code == 409:
                    time.sleep(0.1 * (attempt + 1)) # Backoff
                    continue # Retry
                raise e
            except RuntimeError as e:
                 # Check if our custom wrapper raised it (though we pass 409)
                 if "409" in str(e):
                      time.sleep(0.1 * (attempt + 1))
                      continue
                 raise e

        raise RuntimeError(f"Failed to set property '{original_key}' after {self.MAX_RETRIES} attempts due to conflicts.")

    def get(self, item: str) -> Value:
        """Get a value for a key."""
        page_id = self._find_page_id()
        if page_id is None:
            raise KeyError(item)
        
        hex_key = self._encode_key(item)
        prop = self.__parent._request('GET', f'/rest/api/content/{page_id}/property/{hex_key}')
        
        if not prop:
            raise KeyError(item)
            
        value_data = prop['value']
        # Migration: Check if value is new dict format or old simple structure (if any)
        # We enforce dict format now.
        if isinstance(value_data, dict) and 'value' in value_data and 'mode' in value_data:
            return self.__serializer.deserialize(value_data['value'])
        else:
             # Just in case some other data got there
             raise ValueError(f"Invalid property format for key {item}")

    def delete(self, item: str) -> None:
        """Delete a key."""
        page_id = self._find_page_id()
        if page_id is None:
            return
            
        hex_key = self._encode_key(item)
        
        # Check if exists
        prop = self.__parent._request('GET', f'/rest/api/content/{page_id}/property/{hex_key}')
        if not prop:
            return
            
        try:
            self.__parent._request('DELETE', f'/rest/api/content/{page_id}/property/{hex_key}')
        except RuntimeError:
            pass # Already deleted or error
            
        if self.__sync_page_content:
            self.__sync_content(page_id)

    def drop(self) -> None:
        """Delete the table page."""
        page_id = self._find_page_id()
        if page_id is None:
            return
        self.__parent._request('DELETE', f'/rest/api/content/{page_id}')

    def __sync_content(self, page_id: str) -> None:
        """Update page body to list all properties (Best effort)."""
        try:
            # 1. Fetch all properties with values efficiently
            all_props = {}
            url = f'/rest/api/content/{page_id}/property'
            params = {'limit': 100, 'expand': 'value'}
            
            while url:
                data = self.__parent._request('GET', url, params=params)
                if not data:
                    break
                    
                for prop in data.get('results', []):
                    hex_key = prop['key']
                    try:
                        key = self._decode_key(hex_key)
                        value_data = prop.get('value', {})
                        # Handle value format
                        if isinstance(value_data, dict) and 'value' in value_data:
                            val_str = str(value_data['value'])
                            all_props[key] = val_str
                        else:
                            all_props[key] = str(value_data)
                    except ValueError:
                        pass # Not our property
                
                next_link = data.get('_links', {}).get('next')
                if next_link:
                    url = next_link
                    params = None
                else:
                    url = None

            # 2. Build HTML
            rows = []
            for key, val_str in sorted(all_props.items()):
                if len(val_str) > 100: val_str = val_str[:97] + "..."
                rows.append(f'<tr><th>{html.escape(key)}</th><td>{html.escape(val_str)}</td></tr>')
            
            rows_html = '\n'.join(rows)
            new_content = f'''<p><strong>Content Properties (read-only, do not edit)</strong></p>
<table>
<colgroup><col /><col /></colgroup>
<tbody>
{rows_html}
</tbody>
</table>'''

            # 3. Update page with optimistic locking
            for attempt in range(self.MAX_RETRIES):
                try:
                    page = self.__parent._request('GET', f'/rest/api/content/{page_id}', params={'expand': 'version'})
                    version = page['version']['number'] + 1
                    
                    payload = {
                        'version': {'number': version},
                        'type': 'page',
                        'title': page['title'],
                        'body': {
                            'storage': {
                                'value': new_content,
                                'representation': 'storage',
                            }
                        }
                    }
                
                    self.__parent._request('PUT', f'/rest/api/content/{page_id}', json=payload)
                    return
                except (RuntimeError, requests.exceptions.HTTPError) as e:
                    if "409" in str(e) or (hasattr(e, 'response') and e.response.status_code == 409):
                        time.sleep(0.1 * (attempt+1))
                        continue
                    # Only log warning if sync fails, don't crash main operation
                    print(f"Warning: Sync page update failed (attempt {attempt}): {e}")
        except Exception as e:
            # Sync is optional, log but don't fail the operation
            print(f"Warning: Failed to sync page content: {e}")




import io
import posixpath
from typing import Iterable

try:
    from webdav3.client import Client as WebdavClient
    from webdav3.exceptions import (
        WebDavException,
        RemoteResourceNotFound,
        RemoteParentNotFound,
        MethodNotSupported,
    )
except ImportError:
    raise ImportError('Requires: pip install webdavclient3')

from .mock import DictatureTableMock, DictatureBackendMock, Value, ValueMode, ValueSerializer, ValueSerializerMode


class DictatureBackendWebdav(DictatureBackendMock):
    def __init__(
            self,
            client: WebdavClient,
            dir_prefix: str = 'db_',
            item_prefix: str = 'item_'
    ) -> None:
        self.__client = client
        self.__dir_prefix = dir_prefix
        self.__item_prefix = item_prefix

    def keys(self) -> Iterable[str]:
        try:
            resources = self.__client.list(get_info=True)

            for resource_info in resources:
                name = posixpath.basename(posixpath.dirname(resource_info.get('path')))
                is_dir = resource_info.get('isdir', False)

                if not name:
                    continue

                if is_dir and name.startswith(self.__dir_prefix):
                    table_name_encoded = name[len(self.__dir_prefix):]
                    # noinspection PyProtectedMember
                    yield DictatureTableWebdav._filename_decode(table_name_encoded, suffix='')
        except RemoteResourceNotFound:
            pass

    def table(self, name: str) -> 'DictatureTableMock':
        return DictatureTableWebdav(self.__client, name, self.__dir_prefix, self.__item_prefix)


class DictatureTableWebdav(DictatureTableMock):
    def __init__(self, client: WebdavClient, name: str, db_prefix: str, prefix: str) -> None:
        self.__client = client
        self.__encoded_name = self._filename_encode(name, suffix='')
        self.__path = posixpath.join(db_prefix + self.__encoded_name)
        self.__prefix = prefix
        self.__name_serializer = ValueSerializer(mode=ValueSerializerMode.filename_only)
        self.__value_serializer = ValueSerializer(mode=ValueSerializerMode.any_string)

    def keys(self) -> Iterable[str]:
        try:
            resources = self.__client.list(self.__path, get_info=True)

            for resource_info in resources:
                name = posixpath.basename(resource_info.get('path'))
                is_dir = resource_info.get('isdir', False)

                if not name:
                    continue

                if not is_dir and name.startswith(self.__prefix):
                    item_name_encoded = name[len(self.__prefix):]
                    yield self._filename_decode(item_name_encoded, suffix='.txt')

        except RemoteResourceNotFound:
            pass

    def drop(self) -> None:
        try:
            self.__client.clean(self.__path)
        except RemoteResourceNotFound:
            pass

    def create(self) -> None:
        try:
            if not self.__client.check(self.__path):
                self.__client.mkdir(self.__path)
        except MethodNotSupported as e:
            if not self.__client.check(self.__path):
                raise e

    def set(self, item: str, value: Value) -> None:
        item_path = self.__item_path(item)
        self.create()

        save_data: str = self.__value_serializer.serialize(value)
        data_bytes = save_data.encode('utf-8')

        with io.BytesIO(data_bytes) as buffer:
            try:
                self.__client.upload_to(buffer, item_path)
            except WebDavException as e:
                # If upload fails, ensure the directory exists and retry once
                # This handles race conditions where mkdir hasn't completed yet
                if '409' in str(e) or 'Conflict' in str(e):
                    # Force directory creation and wait for it to exist
                    self.__client.mkdir(self.__path)
                    # Retry the upload
                    buffer.seek(0)  # Reset buffer position
                    self.__client.upload_to(buffer, item_path)
                else:
                    raise

    def get(self, item: str) -> Value:
        item_path = self.__item_path(item)
        buffer = io.BytesIO()
        try:
            self.__client.download_from(buffer, item_path)
            buffer.seek(0)
            save_data = buffer.read().decode('utf-8')
        except RemoteResourceNotFound:
            raise KeyError(item)
        return self.__value_serializer.deserialize(save_data)

    def delete(self, item: str) -> None:
        """
        Deletes the file corresponding to the item. No error if not found.
        Raises WebDavException for other deletion errors.
        """
        item_path = self.__item_path(item)
        try:
            self.__client.clean(item_path)
        except RemoteResourceNotFound:
            pass

    def __item_path(self, item: str) -> str:
        encoded_item_name = self.__prefix + self._filename_encode(item, suffix='.txt')
        full_path = posixpath.join(self.__path, encoded_item_name)
        return full_path

    @staticmethod
    def _filename_encode(name: str, suffix: str = '.txt') -> str:
        return ValueSerializer(mode=ValueSerializerMode.filename_only).serialize(Value(
            value=name,
            mode=ValueMode.string.value
        )) + suffix

    @staticmethod
    def _filename_decode(name: str, suffix: str = '.txt') -> str:
        if suffix and name.endswith(suffix):
            name = name[:-len(suffix)]
        return ValueSerializer(mode=ValueSerializerMode.filename_only).deserialize(name).value



"""Backend that wraps a DictatureTable and uses prefixes to simulate multiple tables."""
from typing import Iterable

from .mock import DictatureBackendMock, DictatureTableMock, Value, ValueMode


class DictatureSingleTableBackend(DictatureBackendMock):
    """
    Backend that stores all tables in a single DictatureTable using prefixes.
    
    This allows using any existing DictatureTable as a backend for multiple
    virtual tables, distinguished by key prefixes.
    """

    def __init__(self, table: 'DictatureTable', separator: str = '::') -> None:
        """
        Create a new SingleTableBackend.
        :param table: The underlying DictatureTable to store all data in
        :param separator: String used to separate table name from key (default: '::')
        """
        self.__table = table
        self.__separator = separator

    def keys(self) -> Iterable[str]:
        """Return all virtual table names."""
        seen = set()
        for key in self.__table.keys():
            if self.__separator in key:
                table_name = key.split(self.__separator, 1)[0]
                if table_name not in seen:
                    seen.add(table_name)
                    yield table_name

    def table(self, name: str) -> 'DictatureSingleTableTable':
        return DictatureSingleTableTable(self.__table, name, self.__separator)


class DictatureSingleTableTable(DictatureTableMock):
    """Virtual table within a SingleTableBackend."""
    
    def __init__(self, parent: 'DictatureTable', table_name: str, separator: str) -> None:
        self.__parent = parent
        self.__prefix = table_name + separator

    def keys(self) -> Iterable[str]:
        """Return all keys in this virtual table."""
        for key in self.__parent.keys():
            if key.startswith(self.__prefix):
                yield key[len(self.__prefix):]

    def drop(self) -> None:
        """Delete all keys for this virtual table."""
        # Collect keys first to avoid modifying during iteration
        keys_to_delete = list(self.keys())
        for key in keys_to_delete:
            del self.__parent[self.__prefix + key]

    def create(self) -> None:
        pass  # Already created by the parent

    def set(self, item: str, value: Value) -> None:
        """Set a value for a key in this virtual table."""
        self.__parent[self.__prefix + item] = {'value': value.value, 'mode': value.mode}

    def get(self, item: str) -> Value:
        """Get a value for a key in this virtual table."""
        value = self.__parent[self.__prefix + item]
        return Value(value['value'], ValueMode(value['mode']))

    def delete(self, item: str) -> None:
        """Delete a key from this virtual table."""
        del self.__parent[self.__prefix + item]
